package com.fossgalaxy.games.tbs.editor.resources;

import com.fossgalaxy.games.tbs.editor.EditorView.EditorModel;
import com.fossgalaxy.games.tbs.io.map2.MapDef;
import com.fossgalaxy.games.tbs.parameters.GameSettings;
import com.fossgalaxy.games.tbs.parameters.ResourceType;

import javax.swing.table.AbstractTableModel;
import java.util.ArrayList;
import java.util.List;

public class ResourceModel extends AbstractTableModel {
    private final String[] columns = {"", "P1", "P2"};

    private EditorModel model;
    private List<ResourceType> types;
    private GameSettings settings;

    public ResourceModel(EditorModel model, GameSettings settings){
        this.model = model;
        this.settings = settings;
        this.types = new ArrayList<>(settings.getResouceTypes());
    }

    @Override
    public int getRowCount() {
        return types.size();
    }

    @Override
    public int getColumnCount() {
        return columns.length;
    }

    @Override
    public String getColumnName(int column) {
        return columns[column];
    }

    @Override
    public Class<?> getColumnClass(int columnIndex) {
        if (columnIndex == 0) {
            return String.class;
        }

        return Integer.class;
    }

    @Override
    public boolean isCellEditable(int rowIndex, int columnIndex) {
        return columnIndex != 0;
    }

    @Override
    public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
        if (columnIndex == 0){
            return;
        }

        MapDef def = model.getDef();
        if (def == null) {
            return;
        }

        ResourceType resourceType = types.get(rowIndex);
        int intVal = (Integer)aValue;
        def.setStartingResource(columnIndex-1, resourceType.getName(), intVal);
        fireTableCellUpdated(rowIndex, columnIndex);
    }

    @Override
    public Object getValueAt(int row, int col) {

        ResourceType resourceType = types.get(row);

        if (col == 0){
            return resourceType.getName();
        }

        MapDef def = model.getDef();
        if (def == null) {
            return 0;
        }

        return (Integer) def.getStartingResource(col - 1, resourceType.getName());
    }

}
